"use strict";
// node docbuilder.js --mode html --no-header --no-footer -i modules.yml 
const
    fs = require('fs'),
    hljs = require('highlight.js'),
    yaml = require('yamljs'),
    underscore = require('underscore'),
    format = require("string-template"),
    marked = require('marked'),
    config = require('./docbuilder_args_parser.js');

const TYPE_MAP = 4;
const langs = ["en", "ru"];
const messages = {
    constants: {en: "Constants", ru: "Константы"},
    functions: {en: "Functions", ru: "Функции"},
    types: {en: "Types", ru: "Типы"},
    example: {en: "Example", ru: "Пример"}
};

// Configure marked to deal with headers
const markedRenderer = new marked.Renderer();
markedRenderer.heading = (text, level) => {
    const br = (level <= 2) ? '<br/>\n' : '';
    const anchor = /{#([\w\s\-]+)}/.exec(text);
    if (anchor == null) {
        return br + '<h' + level + '>' + text + '</h' + level + '>';
    }
    const head = text.substring(0, anchor.index - 1);
    return br + '<h' + level + ' id="' + anchor[1] + '"' + '>' +
            head + '</h' + level + '>';
};
marked.setOptions({
    highlight: code => hljs.highlight(config.codelang, code).value
});

// Headers / footers
const content = {header: { }, footer: { }};
underscore.each(langs, lang => {
    let ext = config.mode === config.MODE_HTML ? '.html' : '.md';
    let header = '';
    if (config.useHeader)
        header = fs.readFileSync(config.outPath + '/header_' + lang + ext, 'utf8');
    content.header[lang] = header;
    let footer = '';
    if (config.useFooter)
        footer = fs.readFileSync(config.outPath + '/footer_' + lang + ext, 'utf8');
    content.footer[lang] = footer;
});

// Create directories
mkdirSync(config.outPath);
if (config.mode == config.MODE_MARKDOWN) {
    underscore.each(langs, lang => {
        mkdirSync(config.outPath + '/' + lang);
        mkdirSync(config.outPath + '/' + lang + '/modules');
    });
}

// Load YML
let modules = yaml.load(config.inPath);
modules = underscore.filter(modules, config.filterScope);
switch (config.mode) {

    case config.MODE_MARKDOWN:
        underscore.each(modules, function(module) {
            underscore.each(langs, function(lang) {
                processMarkdown(lang, module);
            });
        });
        break;
        
    case config.MODE_MARKDOWN_SINGLE:
        underscore.each(langs, function(lang) {
            const path = config.outPath + "/modules_" + lang;
            const file = fs.createWriteStream(path + '.md');
            file.write(content.header[lang]);
            underscore.each(modules, function(module) {
                processMarkdownSingle(file, lang, module);
            });
            file.write(content.footer[lang]);
            file.end();
            console.log(path + '.md generated.');
        });
        break;
        
    case config.MODE_HTML:
        underscore.each(langs, function(lang) {
            const path = config.outPath + "/modules_" + lang;
            fs.readFile(path + '.md', 'utf8', function (err, data) {
                if (err) return console.log(err);
                const file = fs.createWriteStream(path + '.html');
                file.write(content.header[lang]);
                file.write(marked(data, {renderer: markedRenderer}));
                file.write(content.footer[lang]);
                file.end();
                console.log(path + '.html generated.');
            });
        });
        break;
}

function processMarkdown(lang, module) {
    const path = config.outPath + "/" + lang + "/modules/" + module.name + getSuffix(module.scope);
    const file = fs.createWriteStream(path + '.md');
    file.write(content.header[lang]);
    processMarkdownModule(file, lang, module);
    file.write(content.footer[lang]);
    file.end();
    console.log(path + '.md generated.');
}

function processMarkdownSingle(file, lang, module) {
    file.write('\n\n');
    processMarkdownModule(file, lang, module);
    file.write('\n');
}

function processMarkdownModule(file, lang, module) {
    file.write(format('## {name} {#{anchorname}}\n\n', {
        name: module.name,
        anchorname: module.name + getSuffix(module.scope)
    }));
    file.write(format('{desc}\n', {desc: getValue(module, 'desc', lang)}));

    // Constants
    const filteredConstants = underscore.filter(module.constants, config.filterScope);
    if (filteredConstants.length > 0) {
        file.write(format('\n\n### {constants}\n', {constants: messages.constants[lang]}));
        underscore.each(filteredConstants, constant => {
            file.write(format('\n`{name}` : *{typeName}*', constant));
            writeScope(file, constant.scope);

            file.write(' = ');
            const constValue = getValue(constant, 'value', lang);
            if (constant.type !== TYPE_MAP) {
                file.write(format('`{value}`', {value: constValue}));
            } else {
                const mapValues = constValue.replace(/[\{\}]/g, '').split(', ');
                if (mapValues.length >= 7) {
                    file.write('\n\n```\n{\n  ');
                    file.write(mapValues.join(",\n  "));
                    file.write('\n}\n```');
                } else {
                    file.write(format('`{value}`', {value: constValue}));
                }
            }
            file.write('\n');

            var desc = getValue(constant, 'desc', lang);
            if (!isEmpty(desc)) {
                file.write(format('\n{desc}\n', { desc }));
            }
        });
    }

    writeFunctions(file, module.functions, lang);

    // Types
    const filteredTypes = underscore.filter(module.types, config.filterScope);
    if (filteredTypes.length > 0) {
        file.write(format('\n\n### {types}\n', {types: messages.types[lang]}));
        underscore.each(filteredTypes, type => {
            file.write(format('\n`{name}`', type));
            writeScope(file, type.scope);

            const desc = getValue(type, 'desc', lang);
            if (typeof desc !== "undefined") {
                file.write(format('{desc}\n', { desc }));
            }

            writeFunctions(file, type.functions, lang);
            file.write('\n');
        });
    }
}

function writeFunctions(file, functions, lang) {
    const filteredFunctions = underscore.filter(functions, config.filterScope);
    if (filteredFunctions.length === 0) return;
    
    file.write(format('\n\n### {functions}\n', {functions: messages.functions[lang]}));
    underscore.each(filteredFunctions, func => {
        file.write(format('\n`{name}({args})`', func));
        writeScope(file, func.scope);

        const desc = getValue(func, 'desc', lang);
        if (!isEmpty(desc)) {
            file.write(format(' - {desc}\n', { desc }));
        } else {
            file.write('\n');
        }

        const example = getValue(func, 'example', lang);
        if (typeof example !== "undefined") {
            file.write(format('\n{example}\n\n```{codelang}\n{code}\n```\n', {
                example: messages.example[lang],
                code: example,
                codelang: config.codelang
            }));
        }
    });
}

function writeScope(file, scope) {
    if (isEmpty(config.scope)) {
        switch (scope) {
            case "android":
                file.write(' *Android*');
                break;
            case "desktop":
                file.write(' *Desktop*');
                break;
        }
    }
}

function getValue(object, key, lang) {
    if (lang != 'en') {
        const newKey = (key + '_' + lang);
        if (newKey in object) {
            key = newKey;
        }
    }
    return object[key];
}

function getSuffix(scope) {
    if (isEmpty(config.scope) && scope === 'android')  {
        return '_' + scope;
    }
    return '';
}

function mkdirSync(path) {
    try {
        fs.mkdirSync(path);
    } catch(e) {
        if ( e.code != 'EEXIST' ) throw e;
    }
}

function isEmpty(s) {
    return (typeof s == 'undefined') || s.length == 0;
}